from typing import Literal, Optional, Sequence, Union

import aws_cdk as cdk
import constructs
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_iam as iam
from aws_cdk import aws_s3 as s3

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstructMixins


class EnvBaseBucket(s3.Bucket, EnvBaseConstructMixins):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        bucket_name: Optional[str],
        removal_policy: cdk.RemovalPolicy = cdk.RemovalPolicy.RETAIN,
        account_id: str = cdk.Aws.ACCOUNT_ID,
        region: str = cdk.Aws.REGION,
        lifecycle_rules: Optional[Sequence[s3.LifecycleRule]] = None,
        inventories: Optional[Sequence[s3.Inventory]] = None,
        auto_delete_objects: bool = False,
        bucket_key_enabled: bool = False,
        block_public_access: Optional[s3.BlockPublicAccess] = s3.BlockPublicAccess.BLOCK_ALL,
        public_read_access: bool = False,
        **kwargs,
    ):
        self.env_base = env_base
        self._full_bucket_name = bucket_name
        if bucket_name is not None:
            self._full_bucket_name = env_base.get_bucket_name(
                base_name=bucket_name, account_id=account_id, region=region
            )
        super().__init__(
            scope,
            id,
            access_control=s3.BucketAccessControl.PRIVATE,
            auto_delete_objects=auto_delete_objects,
            block_public_access=block_public_access,
            bucket_key_enabled=bucket_key_enabled,
            bucket_name=self.bucket_name,
            public_read_access=public_read_access,
            removal_policy=removal_policy,
            lifecycle_rules=lifecycle_rules,
            inventories=inventories,
            **kwargs,
        )

    @property
    def bucket_name(self) -> str:
        return self._full_bucket_name or super().bucket_name

    def grant_permissions(
        self,
        role: Optional[iam.IRole],
        *permissions: Literal["rw", "r", "w", "d"],
    ):
        grant_bucket_access(self, role, *permissions)


def grant_bucket_access(
    bucket: Union[s3.Bucket, Sequence[s3.Bucket]],
    role: Optional[iam.IRole],
    *permissions: Literal["rw", "r", "w", "d"],
):
    if not role:
        return
    for bucket in [bucket] if isinstance(bucket, s3.Bucket) else bucket:
        for bucket_permission in permissions:
            if bucket_permission == "rw":
                bucket.grant_read_write(role)
            elif bucket_permission == "r":
                bucket.grant_read(role)
            elif bucket_permission == "w":
                bucket.grant_write(role)
            elif bucket_permission == "d":
                bucket.grant_delete(role)



from typing import Literal, Optional, cast

import aws_cdk as cdk


def build_arn(
    partition: str = "aws",
    service: Optional[str] = None,
    region: Optional[str] = None,
    account: Optional[str] = None,
    resource_id: Optional[str] = None,
    resource_type: Optional[str] = None,
    resource_delim: Literal["/", ":"] = ":",
) -> str:
    service = service or "*"
    region = region if region is not None else cast(str, cdk.Aws.REGION)
    account = account if account is not None else cast(str, cdk.Aws.ACCOUNT_ID)
    resource_id = resource_id or "*"

    root_arn = f"arn:{partition}:{service}:{region}:{account}"
    if resource_type is not None:
        return f"{root_arn}:{resource_type}{resource_delim}{resource_id}"
    else:
        return f"{root_arn}:{resource_id}"


def build_batch_arn(
    region: Optional[str] = None,
    account: Optional[str] = None,
    resource_id: Optional[str] = None,
    resource_type: Optional[
        Literal["compute-environment", "job", "job-definition", "job-queue"]
    ] = None,
) -> str:
    return build_arn(
        service="batch",
        region=region,
        account=account,
        resource_id=resource_id,
        resource_type=resource_type,
        resource_delim="/",
    )


def build_dynamodb_arn(
    region: Optional[str] = None,
    account: Optional[str] = None,
    resource_id: Optional[str] = None,
    resource_type: Optional[Literal["table"]] = None,
) -> str:
    return build_arn(
        service="dynamodb",
        region=region,
        account=account,
        resource_id=resource_id,
        resource_type=resource_type,
        resource_delim="/",
    )


def build_ecr_arn(
    region: Optional[str] = None,
    account: Optional[str] = None,
    resource_id: Optional[str] = None,
    resource_type: Optional[Literal["repository"]] = None,
) -> str:
    return build_arn(
        service="ecr",
        region=region,
        account=account,
        resource_id=resource_id,
        resource_type=resource_type,
        resource_delim="/",
    )


def build_sfn_arn(
    region: Optional[str] = None,
    account: Optional[str] = None,
    resource_id: Optional[str] = None,
    resource_type: Optional[Literal["*", "activity", "execution", "stateMachine"]] = None,
) -> str:
    return build_arn(
        service="states",
        region=region,
        account=account,
        resource_id=resource_id,
        resource_type=resource_type,
        resource_delim=":",
    )


def build_lambda_arn(
    region: Optional[str] = None,
    account: Optional[str] = None,
    resource_id: Optional[str] = None,
    resource_type: Optional[Literal["function", "event-source-mapping", "layer"]] = None,
) -> str:
    return build_arn(
        service="lambda",
        region=region,
        account=account,
        resource_id=resource_id,
        resource_type=resource_type,
        resource_delim=":",
    )


def build_s3_arn(
    region: Optional[str] = None,
    account: Optional[str] = None,
    resource_id: Optional[str] = None,
    resource_type: Optional[Literal["bucket", "object", "accesspoint", "job"]] = None,
) -> str:
    # https://docs.aws.amazon.com/AmazonS3/latest/userguide/list_amazons3.html#amazons3-resources-for-iam-policies
    # See table above to see why resource type is set to None
    if resource_type in ["bucket", "object"]:
        resource_type = None
        # ARNs for buckets and objects CANNOT have REGION information
        region = ""
        account = ""

    return build_arn(
        service="s3",
        region=region,
        account=account,
        resource_id=resource_id,
        resource_type=resource_type,
        resource_delim=":",
    )



import logging
from dataclasses import dataclass
from typing import Any, Literal, Optional, Tuple, TypeVar, Union, cast

import aws_cdk as cdk
import constructs
from aibs_informatics_aws_utils.batch import to_mount_point, to_volume
from aibs_informatics_aws_utils.constants.efs import (
    EFS_ROOT_PATH,
    EFS_SCRATCH_PATH,
    EFS_SHARED_PATH,
    EFS_TMP_PATH,
    EFSTag,
)
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_ec2 as ec2
from aws_cdk import aws_efs as efs
from aws_cdk import aws_iam as iam
from aws_cdk import aws_lambda as lambda_
from aws_cdk.aws_efs import (
    LifecyclePolicy,
    OutOfInfrequentAccessPolicy,
    PerformanceMode,
    ThroughputMode,
)

from aibs_informatics_cdk_lib.common.aws.iam_utils import grant_managed_policies
from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstruct, EnvBaseConstructMixins
from aibs_informatics_cdk_lib.constructs_.sfn.utils import convert_to_sfn_api_action_case

logger = logging.getLogger(__name__)

T = TypeVar("T")


class EnvBaseFileSystem(efs.FileSystem, EnvBaseConstructMixins):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        vpc: ec2.IVpc,
        file_system_name: Optional[str] = None,
        allow_anonymous_access: Optional[bool] = None,
        enable_automatic_backups: Optional[bool] = None,
        encrypted: Optional[bool] = None,
        lifecycle_policy: Optional[LifecyclePolicy] = None,
        out_of_infrequent_access_policy: Optional[OutOfInfrequentAccessPolicy] = None,
        performance_mode: Optional[PerformanceMode] = None,
        removal_policy: cdk.RemovalPolicy = cdk.RemovalPolicy.DESTROY,
        throughput_mode: Optional[ThroughputMode] = ThroughputMode.BURSTING,
        **kwargs,
    ) -> None:
        self.env_base = env_base
        super().__init__(
            scope,
            id,
            vpc=vpc,
            file_system_name=(
                self.get_name_with_env(file_system_name) if file_system_name else None
            ),
            allow_anonymous_access=allow_anonymous_access,
            enable_automatic_backups=enable_automatic_backups,
            encrypted=encrypted,
            lifecycle_policy=lifecycle_policy,
            out_of_infrequent_access_policy=out_of_infrequent_access_policy,
            performance_mode=performance_mode,
            removal_policy=removal_policy,
            throughput_mode=throughput_mode,
            **kwargs,
        )

        self._root_access_point = self.create_access_point(name="root", path=EFS_ROOT_PATH)

    @property
    def root_access_point(self) -> efs.AccessPoint:
        return self._root_access_point

    def create_access_point(
        self, name: str, path: str, *tags: Union[EFSTag, Tuple[str, str]]
    ) -> efs.AccessPoint:
        """Create an EFS access point

        Note:   We use CfnAccessPoint because the AccessPoint construct does not support tagging
                or naming. We use tags to name it.

        Args:
            name (str): name used for construct id
            path (str): access point path
            tags (List[EFSTag]): tags to add to the access point

        Returns:
            efs.AccessPoint: _description_
        """
        ap_tags = [tag if isinstance(tag, EFSTag) else EFSTag(*tag) for tag in tags]
        if not any(tag.key == "Name" for tag in ap_tags):
            ap_tags.insert(0, EFSTag("Name", name))

        cfn_access_point = efs.CfnAccessPoint(
            self.get_stack_of(self),
            self.get_construct_id(name, "cfn-ap"),
            file_system_id=self.file_system_id,
            access_point_tags=[
                efs.CfnAccessPoint.AccessPointTagProperty(key=tag.key, value=tag.value)
                for tag in ap_tags
            ],
            posix_user=efs.CfnAccessPoint.PosixUserProperty(
                gid="0",
                uid="0",
            ),
            root_directory=efs.CfnAccessPoint.RootDirectoryProperty(
                creation_info=efs.CfnAccessPoint.CreationInfoProperty(
                    owner_gid="0",
                    owner_uid="0",
                    permissions="0777",
                ),
                path=path,
            ),
        )
        return efs.AccessPoint.from_access_point_attributes(
            self,
            self.get_construct_id(name, "access-point"),
            access_point_id=cfn_access_point.attr_access_point_id,
            file_system=self,
        )  # type: ignore

    def as_lambda_file_system(
        self, access_point: Optional[efs.AccessPoint] = None
    ) -> lambda_.FileSystem:
        ap = access_point or self.root_access_point
        return lambda_.FileSystem.from_efs_access_point(
            ap=ap,
            # Must start with `/mnt` per lambda regex requirements
            mount_path=f"/mnt/efs",
        )

    def grant_lambda_access(self, resource: lambda_.Function):
        grant_file_system_access(self, resource)


class EFSEcosystem(EnvBaseConstruct):
    def __init__(
        self,
        scope: constructs.Construct,
        id: Optional[str],
        env_base: EnvBase,
        file_system_name: Optional[str],
        vpc: ec2.Vpc,
        efs_lifecycle_policy: Optional[efs.LifecyclePolicy] = None,
    ) -> None:
        """Construct for setting up an EFS file system

        NOTE: If the EFS filesystem is intended to be deployed in efs.ThroughputMode.BURSTING
              it may be counterproductive to set an efs_lifecycle_policy other than None because
              EFS files in IA tier DO NOT count towards burst credit accumulation calculations.
              See: https://docs.aws.amazon.com/efs/latest/ug/performance.html#bursting
        """
        super().__init__(scope, id, env_base)
        self._file_system = EnvBaseFileSystem(
            scope=self,
            id="fs",
            env_base=self.env_base,
            file_system_name=(
                self.get_name_with_env(file_system_name) if file_system_name else None
            ),
            lifecycle_policy=efs_lifecycle_policy,
            out_of_infrequent_access_policy=efs.OutOfInfrequentAccessPolicy.AFTER_1_ACCESS,
            enable_automatic_backups=False,
            throughput_mode=efs.ThroughputMode.BURSTING,
            removal_policy=cdk.RemovalPolicy.DESTROY,
            vpc=vpc,
        )

        self.shared_access_point = self.file_system.create_access_point(
            name="shared", path=EFS_SHARED_PATH
        )
        self.scratch_access_point = self.file_system.create_access_point(
            name="scratch", path=EFS_SCRATCH_PATH
        )
        self.tmp_access_point = self.file_system.create_access_point(name="tmp", path=EFS_TMP_PATH)
        self.file_system.add_tags(cdk.Tag("blah", self.env_base))

    @property
    def file_system(self) -> EnvBaseFileSystem:
        return self._file_system

    @property
    def as_lambda_file_system(self) -> lambda_.FileSystem:
        return self.file_system.as_lambda_file_system()


@dataclass
class MountPointConfiguration:
    file_system: Optional[Union[efs.FileSystem, efs.IFileSystem]]
    access_point: Optional[Union[efs.AccessPoint, efs.IAccessPoint]]
    mount_point: str
    root_directory: Optional[str] = None
    read_only: bool = False

    def __post_init__(self):
        if not self.access_point and not self.file_system:
            raise ValueError("Must provide either file system or access point")
        if (
            self.access_point
            and self.file_system
            and self.access_point.file_system.file_system_id != self.file_system.file_system_id
        ):
            raise ValueError("File system of Access point and file system must be the same")
        if not self.mount_point.startswith("/"):
            raise ValueError("Mount point must start with /")

    @classmethod
    def from_file_system(
        cls,
        file_system: Union[efs.FileSystem, efs.IFileSystem],
        root_directory: Optional[str] = None,
        mount_point: Optional[str] = None,
        read_only: bool = False,
    ) -> "MountPointConfiguration":
        if not root_directory:
            root_directory = "/"
        if not mount_point:
            mount_point = f"/opt/efs/{file_system.file_system_id}"
        return cls(
            mount_point=mount_point,
            file_system=file_system,
            access_point=None,
            root_directory=root_directory,
            read_only=read_only,
        )

    @classmethod
    def from_access_point(
        cls,
        access_point: Union[efs.AccessPoint, efs.IAccessPoint],
        mount_point: Optional[str] = None,
        read_only: bool = False,
    ) -> "MountPointConfiguration":
        if not mount_point:
            mount_point = f"/opt/efs/{access_point.access_point_id}"
        return cls(
            mount_point=mount_point,
            access_point=access_point,
            file_system=None,
            root_directory=None,
            read_only=read_only,
        )

    @property
    def file_system_id(self) -> str:
        if self.access_point:
            return self.access_point.file_system.file_system_id
        elif self.file_system:
            return self.file_system.file_system_id
        else:
            raise ValueError("No file system or access point provided")

    @property
    def access_point_id(self) -> Optional[str]:
        if self.access_point:
            return self.access_point.access_point_id
        return None

    def to_batch_mount_point(self, name: str, sfn_format: bool = False) -> dict[str, Any]:
        mount_point: dict[str, Any] = to_mount_point(
            self.mount_point, self.read_only, source_volume=name
        )  # type: ignore[arg-type]  # typed dict should be accepted
        if sfn_format:
            return convert_to_sfn_api_action_case(mount_point)
        return mount_point

    def to_batch_volume(self, name: str, sfn_format: bool = False) -> dict[str, Any]:
        efs_volume_configuration: dict[str, Any] = {
            "fileSystemId": self.file_system_id,
        }
        if self.access_point:
            efs_volume_configuration["transitEncryption"] = "ENABLED"
            # TODO: Consider adding IAM
            efs_volume_configuration["authorizationConfig"] = {
                "accessPointId": self.access_point.access_point_id,
                "iam": "DISABLED",
            }
        else:
            efs_volume_configuration["rootDirectory"] = self.root_directory or "/"
        volume: dict[str, Any] = to_volume(
            None,
            name=name,
            efs_volume_configuration=efs_volume_configuration,  # type: ignore
        )
        if sfn_format:
            return convert_to_sfn_api_action_case(volume)
        return volume


def create_access_point(
    scope: constructs.Construct,
    file_system: Union[efs.FileSystem, efs.IFileSystem],
    name: str,
    path: str,
    *tags: Union[EFSTag, Tuple[str, str]],
) -> efs.AccessPoint:
    """Create an EFS access point

    Note:   We use CfnAccessPoint because the AccessPoint construct does not support tagging
            or naming. We use tags to name it.

    Args:
        name (str): name used for construct id
        path (str): access point path
        tags (List[EFSTag]): tags to add to the access point

    Returns:
        efs.AccessPoint: _description_
    """
    ap_tags = [tag if isinstance(tag, EFSTag) else EFSTag(*tag) for tag in tags]

    if not any(tag.key == "Name" for tag in ap_tags):
        ap_tags.insert(0, EFSTag("Name", name))

    cfn_access_point = efs.CfnAccessPoint(
        scope,
        f"{name}-cfn-ap",
        file_system_id=file_system.file_system_id,
        access_point_tags=[
            efs.CfnAccessPoint.AccessPointTagProperty(key=tag.key, value=tag.value)
            for tag in ap_tags
        ],
        posix_user=efs.CfnAccessPoint.PosixUserProperty(
            gid="0",
            uid="0",
        ),
        root_directory=efs.CfnAccessPoint.RootDirectoryProperty(
            creation_info=efs.CfnAccessPoint.CreationInfoProperty(
                owner_gid="0",
                owner_uid="0",
                permissions="0777",
            ),
            path=path,
        ),
    )
    return efs.AccessPoint.from_access_point_attributes(
        scope,
        f"{name}-access-point",
        access_point_id=cfn_access_point.attr_access_point_id,
        file_system=file_system,
    )  # type: ignore


def grant_connectable_file_system_access(
    file_system: Union[efs.IFileSystem, efs.FileSystem],
    connectable: ec2.IConnectable,
    permissions: Literal["r", "rw"] = "rw",
):
    file_system.connections.allow_default_port_from(connectable)
    repair_connectable_efs_dependency(file_system, connectable)


def grant_role_file_system_access(
    file_system: Union[efs.IFileSystem, efs.FileSystem],
    role: Optional[iam.IRole],
    permissions: Literal["r", "rw"] = "rw",
):
    grant_managed_policies(role, "AmazonElasticFileSystemReadOnlyAccess")
    if "w" in permissions:
        grant_managed_policies(role, "AmazonElasticFileSystemClientReadWriteAccess")


def grant_grantable_file_system_access(
    file_system: Union[efs.IFileSystem, efs.FileSystem],
    grantable: iam.IGrantable,
    permissions: Literal["r", "rw"] = "rw",
):
    actions = []
    if "w" in permissions:
        actions.append("elasticfilesystem:ClientWrite")
    file_system.grant(grantable, *actions)


def grant_file_system_access(
    file_system: Union[efs.IFileSystem, efs.FileSystem], resource: lambda_.Function
):
    grant_grantable_file_system_access(file_system, resource)
    grant_role_file_system_access(file_system, resource.role)
    grant_connectable_file_system_access(file_system, resource)


def repair_connectable_efs_dependency(
    file_system: Union[efs.IFileSystem, efs.FileSystem], connectable: ec2.IConnectable
):
    """Repairs cyclical dependency between EFS and dependent connectable

    Reusing code written in this comment

    https://github.com/aws/aws-cdk/issues/18759#issuecomment-1268689132

    From the github comment:

        When an EFS filesystem is added to a Lambda Function (via the file_system= param)
        it automatically sets up networking access between the two by adding
        an ingress rule on the EFS security group. However, the ingress rule resource
        gets attached to whichever CDK Stack the EFS security group is defined on.
        If the Lambda Function is defined on a different stack, it then creates
        a circular dependency issue, where the EFS stack is dependent on the Lambda
        security group's ID and the Lambda stack is dependent on the EFS stack's file
        system object.

        To resolve this, we manually remove the ingress rule that gets automatically created
        and recreate it on the Lambda's stack instead.


    Args:
        connectable (ec2.IConnectable): Connectable

    """
    connections = connectable.connections
    # Collect IDs of all security groups attached to the connections
    connection_sgs = {sg.security_group_id for sg in connections.security_groups}
    # Iterate over the security groups attached to EFS
    for efs_sg in file_system.connections.security_groups:
        # Iterate over the security group's child nodes
        for child in efs_sg.node.find_all():
            # If this is an ingress rule with a "source" equal to one of
            # the connections' security groups
            if (
                isinstance(child, ec2.CfnSecurityGroupIngress)
                and child.source_security_group_id in connection_sgs
            ):
                # Try to remove the node (raise an error if removal fails)
                node_id = child.node.id
                if not efs_sg.node.try_remove_child(node_id):
                    raise RuntimeError(f"Could not remove child node: {node_id}")

    # Finally, configure the connection between the connections object
    # and the EFS file system which will define the new ingress rule on
    # the stack defining the connection object instead.
    connections.allow_to_default_port(file_system)



__all__ = [
    "EnvBaseStack",
    "EnvBaseStackMixins",
    "add_stack_dependencies",
    "get_all_stacks",
]

from typing import List, Optional, cast

import aws_cdk as cdk
import constructs
from aibs_informatics_core.env import EnvBase, EnvType

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstructMixins


def get_all_stacks(scope: constructs.Construct) -> List[cdk.Stack]:
    children = scope.node.children
    return [cast(cdk.Stack, child) for child in children if isinstance(child, cdk.Stack)]


def add_stack_dependencies(source_stack: cdk.Stack, dependent_stacks: List[cdk.Stack]):
    """Add dependencies between stacks

    Args:
        source_stack (Stack): target stack on which a dependency is made
        dependent_stacks (List[Stack]): the stacks adding the dependency
    """
    for dependent_stack in dependent_stacks:
        dependent_stack.add_dependency(source_stack)


class EnvBaseStackMixins(EnvBaseConstructMixins):
    pass


class EnvBaseStack(cdk.Stack, EnvBaseStackMixins):
    def __init__(
        self,
        scope: constructs.Construct,
        id: Optional[str],
        env_base: EnvBase,
        env: Optional[cdk.Environment] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            scope,
            id or env_base.get_construct_id(str(self.__class__)),
            env=env,
            **kwargs,
        )
        self.env_base = env_base
        self.add_tags(*self.stack_tags)

    @property
    def aws_region(self) -> str:
        return cdk.Stack.of(self).region

    @property
    def aws_account(self) -> str:
        return cdk.Stack.of(self).account

    @property
    def stack_tags(self) -> List[cdk.Tag]:
        return [
            *self.construct_tags,
            cdk.Tag(key=self.env_base.ENV_BASE_KEY, value=self.env_base),
        ]

    @property
    def removal_policy(self) -> cdk.RemovalPolicy:
        if self.env_base.env_type == EnvType.DEV:
            return cdk.RemovalPolicy.DESTROY
        return cdk.RemovalPolicy.RETAIN



from typing import Any, Optional

import constructs
from aws_cdk import aws_stepfunctions as sfn


class CommonOperation:
    @classmethod
    def merge_defaults(
        cls,
        scope: constructs.Construct,
        id: str,
        defaults: dict[str, Any],
        input_path: str = "$",
        result_path: Optional[str] = None,
    ) -> sfn.Chain:
        """Wrapper chain that merges input with defaults.



        Args:
            scope (constructs.Construct): construct scope
            id (str): identifier for the states created
            defaults (dict[str, Any]): default values to merge with input
            input_path (str, optional): Input path of object to merge. Defaults to "$".
            result_path (Optional[str], optional): result path to store merged results.
                Defaults to whatever input_path is defined as.

        Returns:
            sfn.Chain: the new chain that merges defaults with input
        """
        new_input_path = result_path if result_path is not None else input_path
        init_state = sfn.Pass(
            scope,
            "Merge Defaults",
            parameters={
                "input": sfn.JsonPath.object_at("$"),
                "default": defaults,
            },
        )
        merge = sfn.Pass(
            scope,
            "Merge",
            parameters={
                "merged": sfn.JsonPath.json_merge(
                    sfn.JsonPath.object_at("$.default"), sfn.JsonPath.object_at("$.input")
                ),
            },
            output_path="$.merged",
        )

        parallel = init_state.next(merge).to_single_state(
            id=id, input_path=input_path, result_path=new_input_path
        )
        restructure = sfn.Pass(
            scope,
            f"{id} Restructure",
            input_path=f"{new_input_path}[0]",
            result_path=new_input_path,
        )
        return parallel.next(restructure)

    @classmethod
    def enclose_chainable(
        cls,
        scope: constructs.Construct,
        id: str,
        definition: sfn.IChainable,
        input_path: Optional[str] = None,
        result_path: Optional[str] = None,
    ) -> sfn.Chain:
        """Enclose the current state machine fragment within a parallel state.

        Notes:
            - If input_path is not provided, it will default to "$"
            - If result_path is not provided, it will default to input_path

        Args:
            id (str): an identifier for the parallel state
            input_path (Optional[str], optional): input path for the enclosed state.
                Defaults to "$".
            result_path (Optional[str], optional): result path to put output of enclosed state.
                Defaults to same as input_path.

        Returns:
            sfn.Chain: the new state machine fragment
        """
        if input_path is None:
            input_path = "$"
        if result_path is None:
            result_path = input_path

        chain = (
            sfn.Chain.start(definition)
            if not isinstance(definition, (sfn.Chain, sfn.StateMachineFragment))
            else definition
        )

        if isinstance(chain, sfn.Chain):
            parallel = chain.to_single_state(
                id=f"{id} Enclosure", input_path=input_path, result_path=result_path
            )
        else:
            parallel = chain.to_single_state(input_path=input_path, result_path=result_path)
        definition = sfn.Chain.start(parallel)

        if result_path and result_path != sfn.JsonPath.DISCARD:
            restructure = sfn.Pass(
                scope,
                f"{id} Enclosure Post",
                input_path=f"{result_path}[0]",
                result_path=result_path,
            )
            definition = definition.next(restructure)

        return definition



import re
from functools import reduce
from typing import Any, ClassVar, List, Pattern, Union, cast

import aws_cdk as cdk
from aws_cdk import aws_stepfunctions as sfn


class JsonReferencePath(str):
    """
    str extension with properties that provide *some* functionality for defining JsonPath reference expressions
    More details: https://github.com/json-path/JsonPath

    Primarily supports "$" reference.


    """

    _EXTRA_PERIODS_PATTERN: ClassVar[Pattern[str]] = re.compile(r"[$.]+")
    _PERIOD_PATTERN: ClassVar[Pattern[str]] = re.compile(r"(?<!\$)\.(?!\$)")
    _PREFIX: ClassVar[str] = "$."
    _SUFFIX: ClassVar[str] = ".$"
    _REF: ClassVar[str] = "$"

    def __new__(cls, content: str):
        return super().__new__(cls, cls.sanitize(content))

    def __add__(self, other):
        return JsonReferencePath(super().__add__("." + other))

    def extend(self, *paths: str) -> "JsonReferencePath":
        return cast(JsonReferencePath, reduce(lambda x, y: x + y, [self, *paths]))

    @property
    def as_key(self) -> str:
        """
        Returns reference path as a key. This appends ".$"
        """
        return f"{self}{self._SUFFIX}" if self else self._REF

    @property
    def as_reference(self) -> str:
        """
        Returns reference path as a value. This prepends "$."
        """
        return f"{self._PREFIX}{self}" if self else self._REF

    @property
    def as_jsonpath_string(self) -> str:
        return sfn.JsonPath.string_at(self.as_reference)

    @property
    def as_jsonpath_object(self) -> cdk.IResolvable:
        return sfn.JsonPath.object_at(self.as_reference)

    @property
    def as_jsonpath_json_to_string(self) -> str:
        return sfn.JsonPath.json_to_string(self.as_jsonpath_object)

    @property
    def as_jsonpath_list(self) -> List[str]:
        return sfn.JsonPath.list_at(self.as_reference)

    @property
    def as_jsonpath_number(self) -> Union[int, float]:
        return sfn.JsonPath.number_at(self.as_reference)

    @classmethod
    def sanitize(cls, s: str) -> str:
        """Sanitizes a string to ensure string has non-consecutive periods and not on the edge."""
        return f'{cls._EXTRA_PERIODS_PATTERN.sub(".", s).strip(".")}'

    @classmethod
    def is_reference(cls, s: Any) -> bool:
        return isinstance(s, JsonReferencePath) or isinstance(s, str) and s.startswith("$")

    @classmethod
    def empty(cls) -> "JsonReferencePath":
        return cls("")



import logging
import os
import re
import subprocess
import tempfile
from pathlib import Path
from typing import ClassVar, Optional, Union

from aibs_informatics_core.collections import ValidatedStr
from aibs_informatics_core.utils.file_operations import remove_path

logger = logging.getLogger(__name__)


class GitUrl(ValidatedStr):
    regex_pattern: ClassVar[re.Pattern] = re.compile(
        r"((?:(?:git|ssh|http(?:s)?)(?::\/\/)(?:[\w\.]+)\/|(?:git@(?:[\w\.]+)):)(?:[\w\.-]+)\/(?:[\w\.-]+)(?:\.git)?)(?:(?:\#|@|\/tree\/)([\/\w\.]+))?"
    )

    @property
    def repo_base_url(self) -> str:
        return f"{self.get_match_groups()[0].removesuffix('.git')}.git"

    @property
    def repo_name(self) -> str:
        return os.path.basename(self.repo_base_url.removesuffix(".git"))

    @property
    def ref(self) -> Optional[str]:
        return self.get_match_groups()[-1]


# REPO_URL_PATTERN = re.compile(r"(https:\/\/github\.com\/[a-zA-Z0-9_.-]+\/[a-zA-Z0-9_.-]+(?:\/[a-zA-Z0-9_\/.-]*)?)|(git@github\.com:[a-zA-Z0-9_.-]+\/[a-zA-Z0-9_.-]+\.git(?:\/[a-zA-Z0-9_\/.-]*)?)")


def is_repo_url(url: str) -> bool:
    """
    Checks if a given URL is a valid Git repository URL.

    Args:
        url (str): The URL to check.

    Returns:
    bool: True if the URL is a valid Git repository URL, False otherwise.
    """
    return GitUrl.is_valid(url)


def is_local_repo(repo_path: Union[str, Path]) -> bool:
    """
    Checks if a given path is a local Git repository.

    Args:
    repo_path (Path): The file system path to check.

    Returns:
    bool: True if the path is a local Git repository, False otherwise.
    """
    repo_path = Path(repo_path)
    try:
        subprocess.check_output(["git", "-C", repo_path, "rev-parse", "--is-inside-work-tree"])
        return True
    except subprocess.CalledProcessError:
        return False


def get_commit_hash(repo_url_or_path: Union[str, Path]) -> str:
    """
    Get the commit hash of the HEAD reference of a Git repository.

    Args:
        repo_url_or_path (Union[str, Path]): The URL or local path of the Git repository.

    Returns:
    str: The commit hash of the HEAD reference.
    """
    if isinstance(repo_url_or_path, str) and is_repo_url(repo_url_or_path):
        return get_commit_hash_from_url(repo_url_or_path)
    elif is_local_repo(repo_url_or_path):
        repo_path = Path(repo_url_or_path)
        return get_commit_hash_from_local(repo_path)
    else:
        raise ValueError("The input must be a string or a Path object.")


def get_repo_url_components(repo_url: str) -> tuple[str, Optional[str]]:
    git_url = GitUrl(repo_url)
    return (git_url.repo_base_url, git_url.ref)


def get_commit_hash_from_url(repo_url: str) -> str:
    try:
        url = GitUrl(repo_url)
        branch = url.ref or "HEAD"
        # Use git ls-remote to get the commit hashes of remote heads
        output = (
            subprocess.check_output(["git", "ls-remote", url.repo_base_url, "--branch", branch])
            .decode("utf-8")
            .strip()
        )
        # The first part of the output is the commit hash of the HEAD reference
        commit_hash = output.split("\t")[0]
        return commit_hash
    except subprocess.CalledProcessError as e:
        logger.error(f"An error occurred: {e}")
        raise e


def get_commit_hash_from_local(repo_path: Union[str, Path]) -> str:
    try:
        # Get the latest commit hash
        commit_hash = (
            subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=repo_path)
            .decode("utf-8")
            .strip()
        )
        return commit_hash
    except subprocess.CalledProcessError as e:
        logger.error(f"An error occurred while trying to get the commit hash from local: {e}")
        raise e


def get_repo_name(repo_url_or_path: Union[str, Path]) -> str:
    """
    Get the name of a Git repository from its URL or local path.

    Args:
        repo_url_or_path (Union[str, Path]): The URL or local path of the Git repository.

    Returns:
    str: The name of the repository.
    """
    if isinstance(repo_url_or_path, str) and is_repo_url(repo_url_or_path):
        return GitUrl(repo_url_or_path).repo_name
    elif is_local_repo(repo_url_or_path):
        repo_path = Path(repo_url_or_path)
        try:
            # Get the remote URL of the 'origin' remote (commonly used name for the default remote)
            remote_url = (
                subprocess.check_output(
                    ["git", "config", "--get", "remote.origin.url"], cwd=repo_path
                )
                .decode("utf-8")
                .strip()
            )

            # Strip trailing slashes or .git if present
            remote_url = remote_url.rstrip("/").rstrip(".git")

            # Extract the repository name
            repo_name = os.path.basename(remote_url)

            return repo_name
        except subprocess.CalledProcessError as e:
            logger.error(f"An error occurred: {e}")
            raise e

    else:
        raise ValueError("The input must be a string or a Path object.")


def construct_repo_path(repo_url: str, target_dir: Optional[Union[str, Path]] = None) -> Path:
    target_dir = Path(target_dir) if target_dir else Path(tempfile.gettempdir())

    repo_name = get_repo_name(repo_url)
    repo_commit_hash = get_commit_hash(repo_url)

    target_base_name = f"{repo_name}_{repo_commit_hash}"

    target_repo_path = target_dir / target_base_name

    return target_repo_path


def clone_repo(
    repo_url: str, target_dir: Optional[Union[str, Path]] = None, skip_if_exists: bool = True
) -> Path:
    """Clone a Git repository into a target directory.

    Args:
        repo_url (str): The URL of the Git repository.
        target_dir (Optional[Union[str, Path]], optional): Target prefix to store repo under.
            The repo will be written to a subdirectory. Defaults to None (defaut tmp dir used).
        skip_if_exists (bool, optional): Skip cloning if the target directory already exists.
            If the target directory exists and the commit hash matches, the function will return
            the existing path. Defaults to True.

    Returns:
        Path: The path to the cloned repository.
    """
    target_path = construct_repo_path(repo_url, target_dir)

    if target_path.exists():
        if skip_if_exists and get_commit_hash(target_path) == get_commit_hash(repo_url):
            logger.info(
                f"Skipping cloning of repository as target path already exists: {target_path}"
            )
            return target_path
        else:
            remove_path(target_path)
    try:
        # Clone the repository into the target directory
        base_url, branch = get_repo_url_components(repo_url)
        cmd: list[str] = ["git", "clone", base_url, target_path.as_posix(), "--single-branch"]
        if branch:
            cmd.extend(["--branch", branch])
        subprocess.check_call(cmd)
        return target_path
    except subprocess.CalledProcessError as e:
        logger.error(f"An error occurred while trying to clone the repo: {e}")
        raise e



import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Mapping, Optional, Sequence, Tuple

import aws_cdk as cdk
import constructs
from aibs_informatics_core.utils.decorators import cached_property
from aibs_informatics_core.utils.hashing import generate_path_hash
from aws_cdk import aws_lambda as lambda_
from aws_cdk import aws_s3_assets, aws_s3_deployment

logger = logging.getLogger(__name__)


PYTHON_GLOB_EXCLUDES = [
    f"**/.git/**",
    f"**/*.{{egg,egg-info,pyc,pyo}}",
    f"**/{{.egg,.egg-info,.eggs}}/**",
    f"**/{{.venv,__pycache__,build,dist}}/**",
]

GLOBAL_GLOB_EXCLUDES = [
    "**/.git/**",
    "**/*.egg",
    "**/.egg-info",
    "**/.pyc",
    "**/.pyo",
    "**/.egg/**",
    "**/.egg-info/**",
    "**/.eggs/**",
    "**/.venv/**",
    "**/__pycache__/**",
    "**/build/**",
    "**/dist/**",
]


PYTHON_REGEX_EXCLUDES = [
    r".*(.git)/.*",
    r".*/(.eggs|.venv|__pycache__|build|dist)/.*",
    r".*.(.egg|.egg-info)/.*",
    r".*.(egg|egg-info|pyc|pyo)/.*",
    r".*/test/.*",
]


@dataclass
class CodeAsset:
    asset_name: str
    asset_props: aws_s3_assets.AssetProps
    default_runtime: lambda_.Runtime
    supported_runtimes: Optional[Sequence[lambda_.Runtime]] = None
    environment: Optional[Mapping[str, str]] = None

    def __post_init__(self):
        if not self.supported_runtimes:
            self.supported_runtimes = [self.default_runtime]

        if self.default_runtime.name not in [_.name for _ in self.supported_runtimes]:
            raise ValueError(
                f"default runtime {self.default_runtime.name} not in list of"
                f"supported runtimes {[_.name for _ in self.supported_runtimes]}"
            )

    @cached_property
    def as_source(self) -> aws_s3_deployment.Source:
        return self.get_source()

    @property
    def as_code(self) -> lambda_.AssetCode:
        return self.get_code()

    def get_environment(self, *overrides: Tuple[str, str]) -> Mapping[str, str]:
        environment = {**(self.environment or {})}
        environment.update(overrides)
        return environment

    def get_asset(self, scope: constructs.Construct, id: str) -> aws_s3_assets.Asset:
        return aws_s3_assets.Asset(
            scope,
            id,
            path=self.asset_props.path,
            readers=self.asset_props.readers,
            asset_hash=self.asset_props.asset_hash,
            asset_hash_type=self.asset_props.asset_hash_type,
            bundling=self.asset_props.bundling,
            exclude=self.asset_props.exclude,
            follow_symlinks=self.asset_props.follow_symlinks,
            ignore_mode=self.asset_props.ignore_mode,
        )

    def get_source(self) -> aws_s3_deployment.Source:
        return aws_s3_deployment.Source.asset(
            path=self.asset_props.path,
            readers=self.asset_props.readers,
            asset_hash=self.asset_props.asset_hash,
            asset_hash_type=self.asset_props.asset_hash_type,
            bundling=self.asset_props.bundling,
            exclude=self.asset_props.exclude,
            follow_symlinks=self.asset_props.follow_symlinks,
            ignore_mode=self.asset_props.ignore_mode,
        )  # type: ignore

    def get_source_zip(self, archive_filename: str) -> aws_s3_deployment.Source:
        if not self.asset_props.bundling:
            raise ValueError(
                f"Cannot create nesed zip of source {self.asset_name} with "
                f"asset props = {self.asset_props}, because no bundlings options set!"
            )
        elif not self.asset_props.bundling.command:
            raise ValueError(
                f"Cannot create nesed zip of source {self.asset_name} with "
                f"asset props = {self.asset_props}, because no bundling command set!"
            )
        bundling_command = self.asset_props.bundling.command
        bundling_command[-1] = " && ".join(
            [
                bundling_command[-1],
                "cd /asset-output",
                # Zips everything in the directory, excluding the archive file
                f"zip -r {archive_filename} . -x {archive_filename} -q",
                # deletes everything in the directory, except the archive file
                f"find . ! \\( -name '{archive_filename}' -o -name '.' -o -name '..' \\) -prune -exec rm -rf {{}} +",
            ]
        )
        return aws_s3_deployment.Source.asset(
            path=self.asset_props.path,
            readers=self.asset_props.readers,
            asset_hash=self.asset_props.asset_hash,
            asset_hash_type=self.asset_props.asset_hash_type,
            bundling=cdk.BundlingOptions(
                image=self.asset_props.bundling.image,
                command=bundling_command,
                entrypoint=self.asset_props.bundling.entrypoint,
                environment=self.asset_props.bundling.environment,
                local=self.asset_props.bundling.local,
                output_type=cdk.BundlingOutput.NOT_ARCHIVED,
                user=self.asset_props.bundling.user,
                security_opt=self.asset_props.bundling.security_opt,
                volumes=self.asset_props.bundling.volumes,
                working_directory=self.asset_props.bundling.working_directory,
            ),
            follow_symlinks=self.asset_props.follow_symlinks,
            exclude=self.asset_props.exclude,
            ignore_mode=self.asset_props.ignore_mode,
        )  # type: ignore

    def get_code(self) -> lambda_.AssetCode:
        return lambda_.Code.from_asset(
            path=self.asset_props.path,
            readers=self.asset_props.readers,
            asset_hash=self.asset_props.asset_hash,
            asset_hash_type=self.asset_props.asset_hash_type,
            bundling=self.asset_props.bundling,
            exclude=self.asset_props.exclude,
            follow_symlinks=self.asset_props.follow_symlinks,
            ignore_mode=self.asset_props.ignore_mode,
        )

    @classmethod
    def create_py_code_asset(
        cls,
        path: Path,
        context_path: Optional[Path],
        requirements_file: Optional[Path] = None,
        includes: Optional[Sequence[str]] = None,
        excludes: Optional[Sequence[str]] = None,
        runtime: lambda_.Runtime = lambda_.Runtime.PYTHON_3_11,
        platform: Optional[str] = "linux/amd64",
        environment: Optional[Mapping[str, str]] = None,
    ) -> "CodeAsset":
        """Returns a NEW code asset

        Returns:
            CodeAsset: The code asset
        """

        if context_path is None:
            context_path = path

        # This is relative to the repo root copied in asset props below
        package_path = path.relative_to(context_path)
        full_path = context_path / package_path

        asset_hash = generate_path_hash(
            path=str(context_path.resolve()),
            includes=list(includes) if includes else None,
            excludes=[*(excludes or []), *PYTHON_REGEX_EXCLUDES],
        )
        host_ssh_dir = str(Path.home() / ".ssh")
        asset_props = aws_s3_assets.AssetProps(
            # CDK bundles lambda assets in a docker container. This causes issues for our local
            # path dependencies. In order to resolve the relative local path dependency,
            # we need to specify the path to the root of the repo.
            path=str(context_path.resolve()),
            asset_hash=asset_hash,
            # It is important to exclude files from the git repo, because
            #   1. it effectively makes our caching for assets moot
            #   2. we also don't want to include certain files for size reasons.
            exclude=[
                *PYTHON_GLOB_EXCLUDES,
                "**/cdk.out/",
            ],
            # ignore_mode=cdk.IgnoreMode.GIT,
            bundling=cdk.BundlingOptions(
                image=runtime.bundling_image,
                working_directory=f"/asset-input/{package_path}",
                entrypoint=["/bin/bash", "-c"],
                command=[
                    # This makes the following commands run together as one
                    # WARNING Make sure not to modify {host_ssh_dir} in any way, in this set of commands!
                    " && ".join(
                        [
                            "set -x",
                            # Copy in host ssh keys that are needed to clone private git repos
                            f"cp -r {host_ssh_dir} /root/.ssh",
                            # Useful debug if anything goes wrong with github SSH related things
                            "ssh -vT git@github.com || true",
                            # Must make sure that the package is not installing using --editable mode
                            "python3 -m pip install --upgrade pip --no-cache",
                            f"pip3 install {'-r ' + requirements_file.as_posix() if requirements_file else '.'} --no-cache -t /asset-output",
                            # TODO: remove botocore and boto3 from asset output
                            # Must make asset output permissions accessible to lambda
                            "find /asset-output -type d -print0 | xargs -0 chmod 755",
                            "find /asset-output -type f -print0 | xargs -0 chmod 644",
                        ]
                    ),
                ],
                user="root:root",
                volumes=[
                    cdk.DockerVolume(
                        host_path=host_ssh_dir,
                        container_path=host_ssh_dir,
                    ),
                ],
                platform=platform,
            ),
        )

        return CodeAsset(
            asset_name=os.path.basename(full_path.resolve()),
            asset_props=asset_props,
            default_runtime=runtime,
            environment=environment,
        )



__all__ = [
    "Env",
    "EnvBase",
    "EnvType",
    "CodePipelineSourceConfig",
    "CodePipelineNotificationsConfig",
    "GlobalConfig",
    "StageConfig",
    "PipelineConfig",
    "BaseProjectConfig",
    "ProjectConfig",
    "ConfigProvider",
]

import logging
from pathlib import Path
from typing import Annotated, Dict, Generic, List, MutableMapping, Optional, Type, TypeVar, Union

import yaml
from aibs_informatics_core.collections import DeepChainMap
from aibs_informatics_core.env import ENV_BASE_KEY, ENV_LABEL_KEY, ENV_TYPE_KEY, EnvBase, EnvType
from aibs_informatics_core.models.unique_ids import UniqueID
from aibs_informatics_core.utils.file_operations import find_paths
from aibs_informatics_core.utils.os_operations import expandvars
from aibs_informatics_core.utils.tools.dicttools import remove_null_values
from pydantic import BaseModel, PlainSerializer, PlainValidator, model_validator

UniqueIDType = Annotated[
    UniqueID, PlainValidator(lambda x: UniqueID(x)), PlainSerializer(lambda x: str(x))
]


class EnvVarStr(str):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if v is not None and not isinstance(v, str):
            raise TypeError("string required")
        return expandvars(v, "", False) if v else v

    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()})"


class Env(BaseModel):
    env_type: EnvType
    label: Annotated[Optional[str], PlainValidator(EnvVarStr.validate)] = None
    account: Annotated[Optional[str], PlainValidator(EnvVarStr.validate)] = None
    region: Annotated[Optional[str], PlainValidator(EnvVarStr.validate)] = None

    @property
    def env_name(self) -> str:
        return self.env_type.value

    @property
    def env_base(self) -> EnvBase:
        """Returns <env_name>[-<env_label>]"""
        return EnvBase.from_type_and_label(self.env_type, self.label)

    @property
    def is_configured(self) -> bool:
        return self.account is not None

    def to_env_var_map(self) -> MutableMapping[str, str]:
        return remove_null_values(
            dict(
                [
                    (ENV_TYPE_KEY, self.env_type),
                    (ENV_LABEL_KEY, self.label),
                    (ENV_BASE_KEY, self.env_base),
                ]
            ),  # type: ignore
            in_place=True,
        )


class CodePipelineBuildConfig(BaseModel):
    ssh_key_secret_name: str
    docker_hub_credentials_secret_name: str


class CodePipelineSourceConfig(BaseModel):
    repository: str
    branch: Annotated[str, PlainValidator(EnvVarStr.validate)]
    codestar_connection: Optional[UniqueIDType] = None
    oauth_secret_name: Optional[str] = None

    @model_validator(mode="after")
    @classmethod
    def check_source_config(cls, v):
        if not v.codestar_connection and not v.oauth_secret_name:
            raise ValueError("Either codestar_connection or oauth_secret_name must be set")
        if v.codestar_connection and v.oauth_secret_name:
            logging.warning("Only one of codestar_connection or oauth_secret_name can be set")
        return v


class CodePipelineNotificationsConfig(BaseModel):
    slack_channel_configuration_arn: Optional[str]
    notify_on_failure: bool = False
    notify_on_success: bool = False

    @property
    def notify_on_any(self) -> bool:
        return self.notify_on_failure or self.notify_on_success


class PipelineConfig(BaseModel):
    enable: bool
    build: CodePipelineBuildConfig
    source: CodePipelineSourceConfig
    notifications: CodePipelineNotificationsConfig


class GlobalConfig(BaseModel):
    pipeline_name: str
    stage_promotions: Dict[EnvType, EnvType]


class StageConfig(BaseModel):
    env: Env
    pipeline: Optional[PipelineConfig] = None


DEFAULT_CONFIG_PATH = "configuration/project.yaml"


G = TypeVar("G", bound=GlobalConfig)
S = TypeVar("S", bound=StageConfig)
P = TypeVar("P", bound="BaseProjectConfig")


class BaseProjectConfig(BaseModel, Generic[G, S]):
    global_config: G
    default_config: S
    default_config_overrides: Dict[EnvType, dict]

    @classmethod
    def get_global_config_cls(cls) -> Type[G]:
        return cls.model_fields["global_config"].annotation  # type: ignore

    @classmethod
    def get_stage_config_cls(cls) -> Type[S]:
        return cls.model_fields["default_config"].annotation  # type: ignore

    def get_stage_config(
        self, env_type: Union[str, EnvType], env_label: Optional[str] = None
    ) -> S:
        """Get default config with `EnvType` overrides (and optional env_label override)

        Args:
            env_type (Union[str, EnvType]): The EnvType override for the stage config to
                be retrieved.
            env_label (Optional[str], optional): An optional env_label override for
                the stage config to be retrieved. Defaults to None (no override).

        Raises:
            e: If the stage config model validation fails (or any other error)

        Returns:
            S: A stage config object
        """
        try:
            stage_config = self.get_stage_config_cls().model_validate(
                {
                    **DeepChainMap(
                        self.default_config_overrides[EnvType(env_type)],
                        self.default_config.model_dump(mode="json", exclude_unset=True),
                    ),
                }
            )
        except Exception as e:
            raise e

        if env_label is None:
            return stage_config
        else:
            stage_config.env.label = env_label
            return stage_config

    @classmethod
    def parse_file(cls: Type[P], path: Union[str, Path]) -> P:
        path = Path(path)

        if path.suffix in (".yml", ".yaml"):
            with open(path, "r") as f:
                return cls.model_validate(yaml.safe_load(f))
        return cls.model_validate_json(json_data=path.read_text())

    @classmethod
    def load_config(cls: Type[P], path: Optional[Union[str, Path]] = None) -> P:
        if path is None:
            paths = find_paths(
                Path.cwd(), include_dirs=False, include_files=True, includes=[r".*/project.yaml"]
            )
            assert (
                len(paths) == 1
            ), f"Expected to find exactly one project.yaml file, but found {len(paths)}: {paths}"
            path = paths[0]
        return cls.parse_file(path=path)

    @classmethod
    def load_stage_config(
        cls: Type["BaseProjectConfig[G, S]"],
        env_type: Union[str, EnvType],
        path: Optional[Union[str, Path]] = None,
    ) -> S:
        proj_config = cls.load_config(path)
        return proj_config.get_stage_config(env_type=env_type)


class ProjectConfig(BaseProjectConfig[GlobalConfig, StageConfig]):
    pass


class ConfigProvider:
    @classmethod
    def get_stage_config(
        cls,
        env_type: Union[str, EnvType],
        path: Optional[Union[str, Path]] = None,
        project_config_cls: Type[BaseProjectConfig[G, S]] = ProjectConfig,
    ) -> S:
        proj_config = project_config_cls.load_config(path)
        return proj_config.get_stage_config(env_type=env_type)



import hashlib
import re
from typing import List, Optional, Union

import aws_cdk as cdk
from aibs_informatics_core.env import EnvBase, EnvBaseMixins, EnvType, ResourceNameBaseEnum
from aws_cdk import Stack
from aws_cdk import aws_iam as iam
from constructs import Construct

from aibs_informatics_cdk_lib.common.aws.iam_utils import grant_managed_policies


class EnvBaseConstructMixins(EnvBaseMixins):
    @property
    def is_dev(self) -> bool:
        return self.env_base.env_type == EnvType.DEV

    @property
    def is_test(self) -> bool:
        return self.env_base.env_type == EnvType.TEST

    @property
    def is_prod(self) -> bool:
        return self.env_base.env_type == EnvType.PROD

    @property
    def is_test_or_prod(self) -> bool:
        return self.is_prod or self.env_base.env_type == EnvType.TEST

    @property
    def construct_tags(self) -> List[cdk.Tag]:
        return []

    @property
    def aws_region(self) -> str:
        return cdk.Stack.of(self.as_construct()).region

    @property
    def aws_account(self) -> str:
        return cdk.Stack.of(self.as_construct()).account

    def add_tags(self, *tags: cdk.Tag):
        for tag in tags:
            cdk.Tags.of(self.as_construct()).add(key=tag.key, value=tag.value)

    def normalize_construct_id(
        self, construct_id: str, max_size: int = 64, hash_size: int = 8
    ) -> str:
        if max_size < hash_size:
            raise ValueError(f"max_size must be greater than hash_size: ")

        # Replace special characters with dashes
        string = re.sub(r"\W+", "-", construct_id)

        # Check if the string exceeds the allowed size
        if len(string) > max_size:
            # Generate a hexdigest of the string

            digest = hashlib.sha256(string.encode()).hexdigest()
            digest = digest[:hash_size]
            string = string[: -len(digest)] + digest

        return string

    def get_construct_id(self, *names: str) -> str:
        return self.build_construct_id(self.env_base, *names)

    def get_child_id(self, scope: Construct, *names: str) -> str:
        if scope.node.id.startswith(self.env_base):
            return "-".join(names)
        return self.build_construct_id(self.env_base, *names)

    def get_name_with_env(self, *names: str) -> str:
        return self.env_base.prefixed(*names)

    def get_resource_name(self, name: Union[ResourceNameBaseEnum, str]) -> str:
        if isinstance(name, ResourceNameBaseEnum):
            return name.get_name(self.env_base)
        return self.env_base.get_resource_name(name)

    def get_stack_of(self, construct: Optional[Construct] = None) -> Stack:
        if construct is None:
            construct = self.as_construct()
        return cdk.Stack.of(construct)

    def as_construct(self) -> Construct:
        assert isinstance(self, Construct)
        return self

    @classmethod
    def build_construct_id(cls, env_base: EnvBase, *names: str) -> str:
        return env_base.get_construct_id(*names)  # , cls.__name__)

    # ---------------------------------------------------------------
    #                       Resource Helpers
    # ---------------------------------------------------------------

    # ----------
    #   IAM

    @classmethod
    def add_managed_policies(
        cls, role: Optional[iam.IRole], *managed_policies: Union[str, iam.ManagedPolicy]
    ):
        grant_managed_policies(role, *managed_policies)


class EnvBaseConstruct(Construct, EnvBaseConstructMixins):
    def __init__(self, scope: Construct, id: Optional[str], env_base: EnvBase) -> None:
        super().__init__(scope, id or self.__class__.__name__)
        self.env_base = env_base
        self.add_tags(*self.construct_tags)

    @property
    def construct_id(self) -> str:
        return self.node.id



from abc import abstractmethod
from typing import Any, Dict, List, Mapping, Optional, Sequence, TypeVar, Union, cast

import aws_cdk as cdk
import constructs
from aibs_informatics_core.collections import ValidatedStr
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_iam as iam
from aws_cdk import aws_lambda as lambda_
from aws_cdk import aws_logs as logs_
from aws_cdk import aws_stepfunctions as sfn

from aibs_informatics_cdk_lib.common.aws.core_utils import build_lambda_arn
from aibs_informatics_cdk_lib.common.aws.sfn_utils import JsonReferencePath
from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstructMixins

T = TypeVar("T", bound=ValidatedStr)

F = TypeVar("F", bound="StateMachineFragment")


def create_log_options(
    scope: constructs.Construct,
    id: str,
    env_base: EnvBase,
    removal_policy: Optional[cdk.RemovalPolicy] = None,
    retention: Optional[logs_.RetentionDays] = None,
) -> sfn.LogOptions:
    return sfn.LogOptions(
        destination=logs_.LogGroup(
            scope,
            env_base.get_construct_id(id, "state-loggroup"),
            log_group_name=env_base.get_state_machine_log_group_name(id),
            removal_policy=removal_policy or cdk.RemovalPolicy.DESTROY,
            retention=retention or logs_.RetentionDays.ONE_MONTH,
        )
    )


def create_role(
    scope: constructs.Construct,
    id: str,
    env_base: EnvBase,
    assumed_by: iam.IPrincipal = iam.ServicePrincipal("states.amazonaws.com"),  # type: ignore[assignment]
    managed_policies: Optional[Sequence[Union[iam.IManagedPolicy, str]]] = None,
    inline_policies: Optional[Mapping[str, iam.PolicyDocument]] = None,
    inline_policies_from_statements: Optional[Mapping[str, Sequence[iam.PolicyStatement]]] = None,
    include_default_managed_policies: bool = True,
) -> iam.Role:
    construct_id = env_base.get_construct_id(id, "role")

    if managed_policies is not None:
        managed_policies = [
            iam.ManagedPolicy.from_aws_managed_policy_name(policy)
            if isinstance(policy, str)
            else policy
            for policy in managed_policies
        ]

    if inline_policies is None:
        inline_policies = {}
    if inline_policies_from_statements:
        inline_policies = {
            **inline_policies,
            **{
                name: iam.PolicyDocument(statements=statements)
                for name, statements in inline_policies_from_statements.items()
            },
        }

    return iam.Role(
        scope,
        construct_id,
        assumed_by=assumed_by,  # type: ignore
        managed_policies=[
            *(managed_policies or []),
            *[
                iam.ManagedPolicy.from_aws_managed_policy_name(policy)
                for policy in (
                    [
                        "AWSStepFunctionsFullAccess",
                        "CloudWatchLogsFullAccess",
                        "CloudWatchEventsFullAccess",
                    ]
                    if include_default_managed_policies
                    else []
                )
            ],
        ],
        inline_policies=inline_policies,
    )


class StateMachineMixins(EnvBaseConstructMixins):
    def get_fn(self, function_name: str) -> lambda_.IFunction:
        cache_attr = "_function_cache"
        if not hasattr(self, cache_attr):
            setattr(self, cache_attr, {})
        resource_cache = cast(Dict[str, lambda_.IFunction], getattr(self, cache_attr))
        if function_name not in resource_cache:
            resource_cache[function_name] = lambda_.Function.from_function_arn(
                scope=self.as_construct(),
                id=self.env_base.get_construct_id(function_name, "from-arn"),
                function_arn=build_lambda_arn(
                    resource_type="function",
                    resource_id=self.env_base.get_function_name(function_name),
                ),
            )
        return resource_cache[function_name]

    def get_state_machine_from_name(self, state_machine_name: str) -> sfn.IStateMachine:
        cache_attr = "_state_machine_cache"
        if not hasattr(self, cache_attr):
            setattr(self, cache_attr, {})
        resource_cache = cast(Dict[str, sfn.IStateMachine], getattr(self, cache_attr))
        if state_machine_name not in resource_cache:
            resource_cache[state_machine_name] = sfn.StateMachine.from_state_machine_name(
                scope=self.as_construct(),
                id=self.env_base.get_construct_id(state_machine_name, "from-name"),
                state_machine_name=self.env_base.get_state_machine_name(state_machine_name),
            )
        return resource_cache[state_machine_name]


def create_state_machine(
    scope: constructs.Construct,
    env_base: EnvBase,
    id: str,
    name: Optional[str],
    definition: sfn.IChainable,
    role: Optional[iam.Role] = None,
    logs: Optional[sfn.LogOptions] = None,
    timeout: Optional[cdk.Duration] = None,
) -> sfn.StateMachine:
    return sfn.StateMachine(
        scope,
        env_base.get_construct_id(id),
        state_machine_name=env_base.get_state_machine_name(name) if name else None,
        logs=(
            logs
            or sfn.LogOptions(
                destination=logs_.LogGroup(
                    scope,
                    env_base.get_construct_id(id, "state-loggroup"),
                    log_group_name=env_base.get_state_machine_log_group_name(name or id),
                    removal_policy=cdk.RemovalPolicy.DESTROY,
                    retention=logs_.RetentionDays.ONE_MONTH,
                )
            )
        ),
        role=cast(iam.IRole, role),
        definition_body=sfn.DefinitionBody.from_chainable(definition),
        timeout=timeout,
    )


class StateMachineFragment(sfn.StateMachineFragment):
    @property
    def definition(self) -> sfn.IChainable:
        return self._definition

    @definition.setter
    def definition(self, value: sfn.IChainable):
        self._definition = value

    @property
    def start_state(self) -> sfn.State:
        return self.definition.start_state

    @property
    def end_states(self) -> List[sfn.INextable]:
        return self.definition.end_states

    def enclose(
        self,
        id: Optional[str] = None,
        input_path: Optional[str] = None,
        result_path: Optional[str] = None,
    ) -> sfn.Chain:
        """Enclose the current state machine fragment within a parallel state.

        Notes:
            - If input_path is not provided, it will default to "$"
            - If result_path is not provided, it will default to input_path

        Args:
            id (str): an identifier for the parallel state
            input_path (Optional[str], optional): input path for the enclosed state.
                Defaults to "$".
            result_path (Optional[str], optional): result path to put output of enclosed state.
                Defaults to same as input_path.

        Returns:
            sfn.Chain: the new state machine fragment
        """
        id = id or self.node.id

        if input_path is None:
            input_path = "$"
        if result_path is None:
            result_path = input_path

        chain = (
            sfn.Chain.start(self.definition)
            if not isinstance(self.definition, (sfn.Chain, sfn.StateMachineFragment))
            else self.definition
        )

        if isinstance(chain, sfn.Chain):
            parallel = chain.to_single_state(
                id=f"{id} Enclosure", input_path=input_path, result_path=result_path
            )
        else:
            parallel = chain.to_single_state(input_path=input_path, result_path=result_path)
        definition = sfn.Chain.start(parallel)

        if result_path and result_path != sfn.JsonPath.DISCARD:
            restructure = sfn.Pass(
                self,
                f"{id} Enclosure Post",
                input_path=f"{result_path}[0]",
                result_path=result_path,
            )
            definition = definition.next(restructure)

        return definition


class EnvBaseStateMachineFragment(StateMachineFragment, StateMachineMixins):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
    ) -> None:
        super().__init__(scope, id)
        self.env_base = env_base

    def to_single_state(
        self,
        *,
        prefix_states: Optional[str] = None,
        state_id: Optional[str] = None,
        comment: Optional[str] = None,
        input_path: Optional[str] = None,
        output_path: Optional[str] = "$[0]",
        result_path: Optional[str] = None,
        result_selector: Optional[Mapping[str, Any]] = None,
    ) -> sfn.Parallel:
        return super().to_single_state(
            prefix_states=prefix_states,
            state_id=state_id,
            comment=comment,
            input_path=input_path,
            output_path=output_path,
            result_path=result_path,
            result_selector=result_selector,
        )

    def to_state_machine(
        self,
        state_machine_name: str,
        role: Optional[iam.Role] = None,
        logs: Optional[sfn.LogOptions] = None,
        timeout: Optional[cdk.Duration] = None,
    ) -> sfn.StateMachine:
        if role is None:
            role = create_role(
                self,
                state_machine_name,
                self.env_base,
                managed_policies=self.required_managed_policies,
                inline_policies_from_statements={
                    "default": self.required_inline_policy_statements,
                },
            )
        else:
            for policy in self.required_managed_policies:
                if isinstance(policy, str):
                    policy = iam.ManagedPolicy.from_aws_managed_policy_name(policy)
                role.add_managed_policy(policy)

            for statement in self.required_inline_policy_statements:
                role.add_to_policy(statement)

        return sfn.StateMachine(
            self,
            self.get_construct_id(state_machine_name),
            state_machine_name=self.env_base.get_state_machine_name(state_machine_name),
            logs=logs or create_log_options(self, state_machine_name, self.env_base),
            role=(
                role if role is not None else create_role(self, state_machine_name, self.env_base)
            ),  # type: ignore[arg-type]
            definition_body=sfn.DefinitionBody.from_chainable(self.definition),
            timeout=timeout,
        )

    @property
    def required_managed_policies(self) -> Sequence[Union[iam.ManagedPolicy, str]]:
        return []

    @property
    def required_inline_policy_statements(self) -> Sequence[iam.PolicyStatement]:
        return []


class LazyLoadStateMachineFragment(EnvBaseStateMachineFragment):
    @property
    def definition(self) -> sfn.IChainable:
        try:
            return self._definition
        except AttributeError:
            self.definition = self.build_definition()
            return self.definition

    @definition.setter
    def definition(self, value: sfn.IChainable):
        self._definition = value

    @abstractmethod
    def build_definition(self) -> sfn.IChainable:
        raise NotImplementedError("Must implement")


class TaskWithPrePostStatus(LazyLoadStateMachineFragment):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        task: Optional[sfn.IChainable],
    ) -> None:
        super().__init__(scope, id, env_base)
        self.task = task
        self.task_name = id

        self.raw_task_input_path = JsonReferencePath("$")

    @property
    def task(self) -> sfn.IChainable:
        assert self._task, "Task must be set"
        return self._task

    @task.setter
    def task(self, value: Optional[sfn.IChainable]):
        self._task = value

    @property
    def task_name(self) -> str:
        return self._task_name

    @task_name.setter
    def task_name(self, value: str):
        self._task_name = value

    def build_definition(self) -> sfn.IChainable:
        # Should only evaluate once, otherwise errors for duplicate construct will be raised\
        task__augment_input = self.task__augment_input
        task__status_started = self.task__status_started
        task__status_failed = self.task__status_failed
        task__status_completed = self.task__status_completed
        task__pre_run = self.task__pre_run
        task__post_run = self.task__post_run

        # ---------------------------
        # START DEFINITION
        # ---------------------------
        definition = sfn.Pass(
            self,
            f"{self.task_name} Start",
            parameters={
                "input": self.raw_task_input_path.as_jsonpath_object,
                "context": self.task_context,
            },
        )

        # -------------
        # AUGMENT INPUT
        # -------------
        if task__augment_input:
            definition = definition.next(
                sfn.Chain.start(task__augment_input).to_single_state(
                    f"{self.task_name} Augment Input",
                    result_path=self.task_input_path.as_reference,
                    output_path=f"{self.task_input_path.as_reference}[0]",
                )
            )

        # -------------
        # PRE TASK
        # -------------
        if task__status_started:
            definition = definition.next(
                sfn.Chain.start(task__status_started).to_single_state(
                    f"{self.task_name} Status Started", result_path=sfn.JsonPath.DISCARD
                )
            )
        if task__pre_run:
            definition = definition.next(
                sfn.Chain.start(task__pre_run).to_single_state(
                    f"{self.task_name} Pre Run", result_path=sfn.JsonPath.DISCARD
                )
            )

        # -------------
        # TASK
        # -------------
        task_chain = sfn.Chain.start(self.task)
        task_enclosure = task_chain.to_single_state(
            f"{self.task_name} Run",
            input_path=self.task_input_path.as_reference,
            result_path=self.task_result_path.as_reference,
        )
        # fmt: off
        definition = (
            definition
            .next(task_enclosure)
            .next(
                sfn.Pass(
                    self,
                    f"{self.task_name} Run (Restructure)",
                    input_path=f"{self.task_result_path.as_reference}[0]",
                    result_path=self.task_result_path.as_reference,
                )
            )
        )
        # fmt: on

        # -------------
        # TASK FAILED
        # -------------
        if task__status_failed:
            task_enclosure.add_catch(
                sfn.Chain.start(task__status_failed)
                .to_single_state(
                    f"{self.task_name} Status Failed", result_path=sfn.JsonPath.DISCARD
                )
                .next(
                    sfn.Fail(
                        self,
                        f"{self.task_name} Fail State",
                        cause=f"Task {self.task_name} failed during execution.",
                    )
                ),
                result_path=self.task_error_path.as_reference,
            )

        # -------------
        # POST TASK
        # -------------
        if task__post_run:
            definition = definition.next(
                sfn.Chain.start(task__post_run).to_single_state(
                    f"{self.task_name} Post Run", result_path=sfn.JsonPath.DISCARD
                )
            )
        if task__status_completed:
            definition = definition.next(
                sfn.Chain.start(task__status_completed).to_single_state(
                    f"{self.task_name} Status Complete", result_path=sfn.JsonPath.DISCARD
                )
            )

        definition.next(
            sfn.Pass(self, f"{self.task_name} End", input_path=self.task_result_path.as_reference)
        )

        # ---------------------------
        # COMPLETE DEFINITION
        # ---------------------------
        return definition

    @property
    def task_input_path(self) -> JsonReferencePath:
        return JsonReferencePath("input")

    @property
    def task_context(self) -> Dict[str, Any]:
        return {}

    @property
    def task_result_path(self) -> JsonReferencePath:
        return JsonReferencePath("result")

    @property
    def task_error_path(self) -> JsonReferencePath:
        return JsonReferencePath("error")

    @property
    def task_context_path(self) -> JsonReferencePath:
        return JsonReferencePath("context")

    @property
    def task__augment_input(self) -> Optional[sfn.IChainable]:
        """Run right after the sfn.Pass 'START' of the state machine fragment. Can be used
        to dynamically fill out 'contexts' JsonReferencePath for subsequent lambdas/tasks to use.
        NOTE: Outputs of this chain are MAINTAINED
        """
        return None

    @property
    def task__status_started(self) -> Optional[sfn.IChainable]:
        """Runs right before "task__pre_run" if defined. Otherwise runs right before main
        task executes.
        NOTE: Outputs within this chain get DISCARDED
        """
        return None

    @property
    def task__status_completed(self) -> Optional[sfn.IChainable]:
        """Runs right after "task__post_run" if defined. Otherwise runs right after main
        task is completed.
        NOTE: Outputs within this chain get DISCARDED
        """
        return None

    @property
    def task__status_failed(self) -> Optional[sfn.IChainable]:
        """Runs if main task fails during execution"""
        return None

    @property
    def task__pre_run(self) -> Optional[sfn.IChainable]:
        """Runs right before main task executes.
        NOTE: Outputs within this chain get DISCARDED
        """
        return None

    @property
    def task__post_run(self) -> Optional[sfn.IChainable]:
        """Runs right after main task is completed
        NOTE: Outputs within this chain get DISCARDED
        """
        return None
